#include <asm/tc3162.h>
#include <asm/io.h>
#include <common.h>

typedef enum{
	EFUSE_STATUS_SUCCESS = 0,
	EFUSE_STATUS_FAIL
} EFUSE_STATUS_T;

typedef enum{
	EFUSE_READ_SUCCESS = 0,
	EFUSE_READ_IDX_ERROR,
	EFUSE_READ_READ_ERROR_ADDR,
	EFUSE_READ_TIMEOUT,
	EFUSE_READ_FAIL
} EFUSE_READ_STATUS_T;

typedef enum {
	EFUSE_WRITE_SUCCESS = 0,
	EFUSE_WRITE_IDX_ERROR,
	EFUSE_WRITE_ERROR_ADDR,
	EFUSE_WRITE_TIMEOUT,
	EFUSE_WRITE_FAIL,
	EFUSE_WRITE_REPAIR_TABLE_FULL,
} EFUSE_WRITE_STATUS_T;


typedef struct ef_data {
	unsigned int data[4];
	unsigned int res_data;
} EF_DATA;

#define EF_TOTAL_NUM				(1)
#define EF_TOTAL_ROW				(128)
#define EF_TOTAL_RES_ROW			(24)
#define EF_TOTAL_COL				(32)
#define EF_TOTAL_BITS				(EF_TOTAL_ROW * EF_TOTAL_COL)
#define EF_TOTAL_RES_BITS			(EF_TOTAL_RES_ROW * EF_TOTAL_COL)

#define COL_TO_BYTE(col)			(col << 4)
#define WORD_TO_BIT(word)			(word << 5)
#define BIT_TO_WORD(bit)			(bit >> 5)

EF_DATA ef_data[EF_TOTAL_COL];

#define EFUSE_TEST					(0)

#define CR_EFUSE_BASE				(0x1FBF8200)

#define CR_EFUSE_CFG				(0x04)
#define CR_EFUSE_CTRL				(0x08)
#define CR_EFUSE_RES_WDATA			(0x0C)
#define CR_EFUSE_WDATA0				(0x10)
#define CR_EFUSE_WDATA1				(0x14)
#define CR_EFUSE_WDATA2				(0x18)
#define CR_EFUSE_WDATA3				(0x1C)
#define CR_EFUSE_BIST				(0x20)
#define CR_EFUSE_CFG2				(0x24)
#define CR_EFUSE_CFG3				(0x28)
#define CR_EFUSE_RES_RDATA			(0x2C)
#define CR_EFUSE_RDATA0				(0x30)
#define CR_EFUSE_RDATA1				(0x34)
#define CR_EFUSE_RDATA2				(0x38)
#define CR_EFUSE_RDATA3				(0x3C)

#define EFUSE_ADDR_MASK				(0x1FF)
#define EFUSE_ADDR_SHIFT			(16)
#define EFUSE_MODE_MASK				(0x3)
#define EFUSE_MODE_SHIFT			(6)
#define EFUSE_MODE_READ				(0x1)
#define EFUSE_MODE_WRITE			(0x3)
#define EFUSE_KICK_SHIFT			(30)
#define EFUSE_LDO_MASK				(0xFFFF)
#define EFUSE_SYNC_MASK				(0x3F)
#define EFUSE_BIST_MASK				(0x3F00)
#define EFUSE_BIST_START			(0x01)
#define EFUSE_BIST_DONE				(0x02)

#define MAX_DELAY_US				(1000000)

#define EFUSE_DBG_LV 1

#define EFUSE_DBG_MSG(level, fmt, args...)	{ \
												if(level >= EFUSE_DBG_LV) { \
													printf(fmt, ##args); \
												} \
											}
static int ef_inited = 0;

static EFUSE_READ_STATUS_T ef_get_row_col(unsigned int bit_offset, unsigned int *p_row, unsigned int *p_col)
{
	if(bit_offset >= EF_TOTAL_BITS) {
		return EFUSE_STATUS_FAIL;
	}

	*p_row = bit_offset % EF_TOTAL_ROW;
	*p_col = bit_offset / EF_TOTAL_ROW;

	return EFUSE_STATUS_SUCCESS;
}

static int ef_bist_test(void)
{
	unsigned int delay = MAX_DELAY_US;

	VPint(CR_EFUSE_BASE + CR_EFUSE_BIST) = EFUSE_BIST_START;

	do
	{
		udelay(1);
		delay--;
	} while(!(VPint(CR_EFUSE_BASE + CR_EFUSE_BIST) & EFUSE_BIST_DONE) && delay);

	if (!delay)
	{
		EFUSE_DBG_MSG(1, "BIST TIMEOUT\n");
		return -1;
	}

	return (VPint(CR_EFUSE_BASE + CR_EFUSE_BIST) & EFUSE_BIST_MASK);
}

static EFUSE_READ_STATUS_T ef_read_data(unsigned int byte_offset, EF_DATA *p_data)
{
	unsigned int ef_ctrl = 0;
	unsigned int delay = MAX_DELAY_US;

	ef_ctrl = VPint(CR_EFUSE_BASE + CR_EFUSE_CTRL);

	if (ef_ctrl & (1 << EFUSE_KICK_SHIFT))
	{
		EFUSE_DBG_MSG(1, "BUSY\n");
		return EFUSE_READ_FAIL;
	}

	ef_ctrl &= ~((EFUSE_ADDR_MASK << EFUSE_ADDR_SHIFT) | (EFUSE_MODE_MASK << EFUSE_MODE_SHIFT));

	ef_ctrl |= ((EFUSE_MODE_READ << EFUSE_MODE_SHIFT) | ((byte_offset & EFUSE_ADDR_MASK) << EFUSE_ADDR_SHIFT));

	VPint(CR_EFUSE_BASE + CR_EFUSE_CTRL) = (ef_ctrl | (1 << EFUSE_KICK_SHIFT));

	do
	{
		udelay(1);
		delay--;
	} while((VPint(CR_EFUSE_BASE + CR_EFUSE_CTRL) & (1 << EFUSE_KICK_SHIFT)) && delay);

	if (!delay)
	{
			EFUSE_DBG_MSG(1, "READ TIMEOUT\n");
			return EFUSE_READ_TIMEOUT;
	}

	p_data->data[0] = VPint(CR_EFUSE_BASE + CR_EFUSE_RDATA0);
	p_data->data[1] = VPint(CR_EFUSE_BASE + CR_EFUSE_RDATA1);
	p_data->data[2] = VPint(CR_EFUSE_BASE + CR_EFUSE_RDATA2);
	p_data->data[3] = VPint(CR_EFUSE_BASE + CR_EFUSE_RDATA3);
	p_data->res_data= VPint(CR_EFUSE_BASE + CR_EFUSE_RES_RDATA);

	return EFUSE_READ_SUCCESS;
}

static EFUSE_WRITE_STATUS_T ef_write_data(unsigned int byte_offset, EF_DATA *p_data)
{
	unsigned int ef_ctrl = 0;
	unsigned int delay = MAX_DELAY_US;

	ef_ctrl = VPint(CR_EFUSE_BASE + CR_EFUSE_CTRL);

	if (ef_ctrl & (1 << EFUSE_KICK_SHIFT))
	{
		EFUSE_DBG_MSG(1, "BUSY\n");
		return EFUSE_WRITE_FAIL;
	}

	ef_ctrl &= ~((EFUSE_ADDR_MASK << EFUSE_ADDR_SHIFT) | (EFUSE_MODE_MASK << EFUSE_MODE_SHIFT));

	ef_ctrl |= ((EFUSE_MODE_WRITE<< EFUSE_MODE_SHIFT) | ((byte_offset & EFUSE_ADDR_MASK) << EFUSE_ADDR_SHIFT));

	VPint(CR_EFUSE_BASE + CR_EFUSE_WDATA0) = p_data->data[0];
	VPint(CR_EFUSE_BASE + CR_EFUSE_WDATA1) = p_data->data[1];
	VPint(CR_EFUSE_BASE + CR_EFUSE_WDATA2) = p_data->data[2];
	VPint(CR_EFUSE_BASE + CR_EFUSE_WDATA3) = p_data->data[3];
	VPint(CR_EFUSE_BASE + CR_EFUSE_RES_WDATA) = p_data->res_data;

	VPint(CR_EFUSE_BASE + CR_EFUSE_CTRL) = (ef_ctrl | (1 << EFUSE_KICK_SHIFT));

	do
	{
		udelay(1);
		delay--;
	} while((VPint(CR_EFUSE_BASE + CR_EFUSE_CTRL) & (1 << EFUSE_KICK_SHIFT)) && delay);

	if (!delay)
	{
			EFUSE_DBG_MSG(1, "WRITE TIMEOUT\n");
			return EFUSE_WRITE_TIMEOUT;
	}

	return EFUSE_WRITE_SUCCESS;
}

static EFUSE_STATUS_T ef_config_ldo(unsigned char on, unsigned char off)
{
	unsigned int ef_cfg = 0;

	if (VPint(CR_EFUSE_BASE + CR_EFUSE_CTRL) & (1 << EFUSE_KICK_SHIFT))
			return EFUSE_STATUS_FAIL;

	ef_cfg = VPint(CR_EFUSE_BASE + CR_EFUSE_CFG);
	ef_cfg &= ~(EFUSE_LDO_MASK);
	ef_cfg |= ((off << 8) | on);

	VPint(CR_EFUSE_BASE + CR_EFUSE_CFG) = ef_cfg;
	EFUSE_DBG_MSG(1, "CR_EFUSE_CFG %lx\n", VPint(CR_EFUSE_BASE + CR_EFUSE_CFG));

	return EFUSE_STATUS_SUCCESS;
}

static EFUSE_STATUS_T ef_config_sync(unsigned char cnt)
{
	unsigned int ef_cfg2 = 0;

	if (VPint(CR_EFUSE_BASE + CR_EFUSE_CTRL) & (1 << EFUSE_KICK_SHIFT))
			return EFUSE_STATUS_FAIL;

	ef_cfg2 = VPint(CR_EFUSE_BASE + CR_EFUSE_CFG2);
	ef_cfg2 &= ~(EFUSE_SYNC_MASK);
	ef_cfg2 |= cnt;

	VPint(CR_EFUSE_BASE + CR_EFUSE_CFG2) = ef_cfg2;
	EFUSE_DBG_MSG(1, "CR_EFUSE_CFG2 %lx\n", VPint(CR_EFUSE_BASE + CR_EFUSE_CFG2));

	return EFUSE_STATUS_SUCCESS;
}

static EFUSE_STATUS_T ef_config_wstb(unsigned short on, unsigned short off)
{
	if (VPint(CR_EFUSE_BASE + CR_EFUSE_CTRL) & (1 << EFUSE_KICK_SHIFT))
			return EFUSE_STATUS_FAIL;

	VPint(CR_EFUSE_BASE + CR_EFUSE_CFG3) = ((off << 16) | on);
	EFUSE_DBG_MSG(1, "CR_EFUSE_CFG3 %lx\n", VPint(CR_EFUSE_BASE + CR_EFUSE_CFG3));

	return EFUSE_STATUS_SUCCESS;
}

static EFUSE_STATUS_T ef_init(unsigned char ldo_on, unsigned char ldo_off, unsigned char sync_cnt, unsigned short wstb_on, unsigned short wstb_off)
{
	unsigned int col = 0;
	EFUSE_READ_STATUS_T status = EFUSE_READ_SUCCESS;

	memset(ef_data, 0, (sizeof(EF_DATA) * EF_TOTAL_COL));

	if (ldo_on || ldo_off)
		ef_config_ldo(ldo_on, ldo_off);

	if (sync_cnt)
		ef_config_sync(sync_cnt);

	if (wstb_on || wstb_off)
		ef_config_wstb(wstb_on, wstb_off);

	for(col = 0; col < EF_TOTAL_COL; col++)
	{
		status = ef_read_data(COL_TO_BYTE(col), &(ef_data[col]));

		if (status != EFUSE_READ_SUCCESS)
		{
			EFUSE_DBG_MSG(1, "ef_init failed: %x \n", status);
			return EFUSE_STATUS_FAIL;
		}
	}

	return EFUSE_STATUS_SUCCESS;
}

static EFUSE_STATUS_T ef_read_word(unsigned int word_offset, unsigned int *p_buf)
{
	unsigned int row = 0;
	unsigned int col = 0;
	unsigned int bit_offset = WORD_TO_BIT((word_offset & 0x7F));
	unsigned char word_idx = 0;

	if (ef_get_row_col(bit_offset, &row, &col))
		return EFUSE_STATUS_FAIL;

	if (row & 0x1F)
		return EFUSE_STATUS_FAIL;

	word_idx = BIT_TO_WORD(row);

	memcpy(p_buf, &(ef_data[col].data[word_idx]), sizeof(unsigned int));

	return EFUSE_STATUS_SUCCESS;
}

static EFUSE_STATUS_T ef_write_word(unsigned int word_offset, unsigned int *p_buf)
{
	EF_DATA wdata;
	unsigned int row = 0;
	unsigned int col = 0;
	unsigned int bit_offset = WORD_TO_BIT((word_offset & 0x7F));
	unsigned char word_idx = 0;
	EFUSE_WRITE_STATUS_T status = EFUSE_WRITE_SUCCESS;

	memset(&wdata, 0, sizeof(EF_DATA));

	if (ef_get_row_col(bit_offset, &row, &col))
		return EFUSE_STATUS_FAIL;

	if (row & 0x1F)
		return EFUSE_STATUS_FAIL;

	word_idx = BIT_TO_WORD(row);

	memcpy(&(wdata.data[word_idx]), p_buf, sizeof(unsigned int));
	wdata.data[word_idx] &= ~(ef_data[col].data[word_idx]);

	status = ef_write_data(COL_TO_BYTE(col), &wdata);

	if (status != EFUSE_WRITE_SUCCESS)
	{
		EFUSE_DBG_MSG(1, "ef_write failed: %x \n", status);
		return EFUSE_STATUS_FAIL;
	}
	else
	{
		ef_data[col].data[word_idx] |= wdata.data[word_idx];
	}

	return EFUSE_STATUS_SUCCESS;
}

#if EFUSE_TEST
typedef struct ef_cfg {
	unsigned char ldo_on;
	unsigned char ldo_off;
	unsigned char sync_cnt;
	unsigned short wstb_on;
	unsigned short wstb_off;
} EF_CFG;

static EF_CFG ef_cfg[10] = {
	{0x01, 0x01, 0x10, 0x4A, 0x0B},
	{0x7F, 0x7F, 0x10, 0x4A, 0x0B},
	{0xFF, 0xFF, 0x10, 0x4A, 0x0B},
	{0x01, 0x01, 0x1F, 0x4A, 0x0B},
	{0x01, 0x01, 0x2F, 0x4A, 0x0B},
	{0x01, 0x01, 0x3F, 0x4A, 0x0B},
	{0x01, 0x01, 0x10, 0x01, 0x01},
	{0x01, 0x01, 0x10, 0xA3, 0x61},
	{0x01, 0x01, 0x10, 0x7FFF, 0x7FFF},
	{0x01, 0x01, 0x10, 0xFFFF, 0xFFFF},
};

static unsigned int data[4][4] = {
	{0xFF0055AA, 0x00FFAA55, 0x55AA00FF, 0x55AA00FF},
	{0x55AAFF00, 0xAA5500FF, 0x00FF55AA, 0xFF00AA55},
	{0x00FFAA55, 0xFF0055AA, 0x55AA00FF, 0x55AA00FF},
	{0xAA5500FF, 0x55AAFF00, 0xFF00AA55, 0x00FF55AA},
};

static int ef_write_and_compare(unsigned int mode, EF_DATA *p_wdata, EF_DATA *p_rdata,
		unsigned int data_pattern, unsigned int check_data, unsigned int res_patten, unsigned int check_res)
{
	unsigned int col = 0, i = 0;

	for(col = 0; col < EF_TOTAL_COL; col++)
	{
		for (i = 0; i < 4; i++)
		{
			if (mode == 2)
			{
				data_pattern = data[(col % 4)][i];
			}

			p_wdata->data[i] = data_pattern;
		}
		p_wdata->res_data = res_patten;

		ef_write_data(COL_TO_BYTE(col), p_wdata);
		ef_read_data(COL_TO_BYTE(col), p_rdata);

		for (i = 0; i < 4; i++)
		{
			if (mode == 2)
			{
				check_data = data[(col % 4)][i];
			}

			if (p_rdata->data[i] != check_data)
			{
				EFUSE_DBG_MSG(1, "ef_test failed: mode %d col %x i %x rdata.data[%x] wdata.data[%x] check_data[%x]\n",
						mode, col, i, p_rdata->data[i], p_wdata->data[i], check_data);
				return -1;
			}
		}

		if (p_rdata->res_data != check_res)
		{
			EFUSE_DBG_MSG(1, "ef_test failed: mode %d col %x rdata.res_data[%x] wdata.res_data[%x] check_res[%x]\n",
					mode, col, p_rdata->res_data, p_wdata->res_data, check_res);
			return -1;
		}
	}

	return 0;
}


static int ef_test(unsigned int mode, unsigned int sub)
{
	EF_DATA wdata;
	EF_DATA rdata;
	unsigned int col = 0, row = 0, i = 0, bit = 0, tmp[4] = {0};
	unsigned int data_pattern = 0;
	unsigned int check_data = 0;
	unsigned int res_patten = 0;
	unsigned int check_res = 0;

	memset(&wdata, 0, sizeof(EF_DATA));
	memset(&rdata, 0, sizeof(EF_DATA));

	switch(mode)
	{
		case 1:
		{
			check_data = data_pattern = 0xFFFFFFFF;
			check_res = res_patten = 0;
			/* Write 0xFFFFFFFF to data and read back to check */
			if (ef_write_and_compare(mode, &wdata, &rdata, data_pattern, check_data, res_patten, check_res))
			{
				return -1;
			}

			data_pattern = 0x0;
			/* Write 0 to data and read back to check */
			if (ef_write_and_compare(mode, &wdata, &rdata, data_pattern, check_data, res_patten, check_res))
			{
				return -1;
			}

			EFUSE_DBG_MSG(1, "ef_test success: mode %d\n", mode);
		}
			break;

		case 2:
		{
			/* Write special pattern to data and read back to check */
			if (ef_write_and_compare(mode, &wdata, &rdata, data_pattern, check_data, res_patten, check_res))
			{
				return -1;
			}

			EFUSE_DBG_MSG(1, "ef_test success: mode %d\n", mode);
		}
			break;

		case 3:
		{
			check_data = data_pattern = 0x0;
			check_res = res_patten = 0xFFFFFF;
			/* Write 0xFFFFFF to res_data and read back to check */
			if (ef_write_and_compare(mode, &wdata, &rdata, data_pattern, check_data, res_patten, check_res))
			{
				return -1;
			}

			res_patten = 0;
			if (ef_write_and_compare(mode, &wdata, &rdata, data_pattern, check_data, res_patten, check_res))
			{
				return -1;
			}

			EFUSE_DBG_MSG(1, "ef_test success: mode %d\n", mode);
		}
			break;

		case 4:
		{
			check_data = data_pattern = 0x0;
			check_res = res_patten = 0x56AA95;
			/* Write special pattern to res_data and read back to check */
			if (ef_write_and_compare(mode, &wdata, &rdata, data_pattern, check_data, res_patten, check_res))
			{
				return -1;
			}

			EFUSE_DBG_MSG(1, "ef_test success: mode %d\n", mode);
		}
			break;

		case 5:
		{
			/* Change timing */
			switch(sub)
			{
				case 1:
				case 2:
				case 3:
				case 4:
				case 5:
				case 6:
				case 7:
				case 8:
				case 9:
					ef_config_ldo(ef_cfg[sub].ldo_on, ef_cfg[sub].ldo_off);
					ef_config_sync(ef_cfg[sub].sync_cnt);
					ef_config_wstb(ef_cfg[sub].wstb_on, ef_cfg[sub].wstb_off);
					break;
				default:
				{
					ef_config_ldo(ef_cfg[0].ldo_on, ef_cfg[0].ldo_off);
					ef_config_sync(ef_cfg[0].sync_cnt);
					ef_config_wstb(ef_cfg[0].wstb_on, ef_cfg[0].wstb_off);

					EFUSE_DBG_MSG(1, "ef_test using default setting\n");
				}
			}

			check_data = data_pattern = 0xFFFFFFFF;
			check_res = res_patten = 0xFFFFFF;
			/* Write 0xFFFFFF and read back to check */
			if (ef_write_and_compare(mode, &wdata, &rdata, data_pattern, check_data, res_patten, check_res))
			{
				return -1;
			}

			EFUSE_DBG_MSG(1, "ef_test success: mode %d\n", mode);
		}
			break;

		case 6:
		{
			if (sub)
			{
				col = (sub - 1);
				wdata.data[(col % 4)] = (1 << col);

				ef_write_data(COL_TO_BYTE(col), &wdata);
				ef_read_data(COL_TO_BYTE(col), &rdata);
			}

			i = ef_bist_test();
			if (sub)
			{
				if (!i)
				{
					EFUSE_DBG_MSG(1, "ef_test failed: mode %d col %x bist_status[%x]\n", mode, col, i);
					return -1;
				}
			}
			else
			{
				if (i)
				{
					EFUSE_DBG_MSG(1, "ef_test failed: mode %d bist_status[%x]\n", mode, i);
					return -1;
				}
			}

			EFUSE_DBG_MSG(1, "ef_test success: mode %d\n", mode);
		}
			break;

		case 7:
		{
			switch(sub)
			{
				case 1:
					col = 0;
					break;
				case 2:
					col = 5;
					break;
				case 4:
					col = 10;
					break;
				case 8:
					col = 15;
					break;
				default:
				{
					col = sub & 0x1F;
					sub = sub & 0xF;
				}
			}

			for (i = 0; i < 4; i++)
				wdata.data[i] = 0xFFFFFFFF;
			wdata.res_data = 0xFFFFFF;

			ef_write_data((COL_TO_BYTE(col) + sub), &wdata);
			ef_read_data((COL_TO_BYTE(col) + sub), &rdata);

			for (i = 0; i < 4; i++)
			{
				if (rdata.data[i] != 0xFFFFFFFF)
				{
					EFUSE_DBG_MSG(1, "ef_test failed: mode %d col %x i %x rdata.data[%x] wdata.data[%x]\n",
							mode, col, i, rdata.data[i], wdata.data[i]);
					return -1;
				}
			}

			if (rdata.res_data != 0xFFFFFF)
			{
				EFUSE_DBG_MSG(1, "ef_test failed: mode %d col %x rdata.res_data[%x] wdata.res_data[%x]\n",
						mode, col, rdata.res_data, wdata.res_data);
				return -1;
			}

			ef_read_data(COL_TO_BYTE(col), &rdata);

			for (i = 0; i < 4; i++)
			{
				if (rdata.data[i] != 0xFFFFFFFF)
				{
					EFUSE_DBG_MSG(1, "ef_test failed: mode %d col %x i %x rdata.data[%x] wdata.data[%x]\n",
							mode, col, i, rdata.data[i], wdata.data[i]);
					return -1;
				}
			}

			if (rdata.res_data != 0xFFFFFF)
			{
				EFUSE_DBG_MSG(1, "ef_test failed: mode %d col %x rdata.res_data[%x] wdata.res_data[%x]\n",
						mode, col, rdata.res_data, wdata.res_data);
				return -1;
			}

			ef_read_data(COL_TO_BYTE((col + 1)), &rdata);

			for (i = 0; i < 4; i++)
			{
				if (rdata.data[i] != 0)
				{
					EFUSE_DBG_MSG(1, "ef_test failed: mode %d col %x i %x rdata.data[%x]\n",
							mode, (col + 1), i, rdata.data[i]);
					return -1;
				}
			}

			if (rdata.res_data != 0)
			{
				EFUSE_DBG_MSG(1, "ef_test failed: mode %d col %x rdata.res_data[%x]\n",
						mode, (col + 1), rdata.res_data);
				return -1;
			}

			EFUSE_DBG_MSG(1, "ef_test success: mode %d\n", mode);
		}
			break;

		case 8:
		{
			check_data = data_pattern = 0xFFFF;
			check_res = res_patten = 0xFFF;
			if (ef_write_and_compare(mode, &wdata, &rdata, data_pattern, check_data, res_patten, check_res))
			{
				return -1;
			}

			data_pattern = 0xFFFF0000;
			check_data = 0xFFFFFFFF;
			res_patten = 0xFFF000;
			check_res = 0xFFFFFF;
			if (ef_write_and_compare(mode, &wdata, &rdata, data_pattern, check_data, res_patten, check_res))
			{
				return -1;
			}

			EFUSE_DBG_MSG(1, "ef_test success: mode %d\n", mode);
		}
			break;

		case 9:
		{
			check_data = data_pattern = 0xAAAAAAAA;
			check_res = res_patten = 0xAAAAAA;
			if (ef_write_and_compare(mode, &wdata, &rdata, data_pattern, check_data, res_patten, check_res))
			{
				return -1;
			}

			data_pattern = 0x55555555;
			check_data = 0xFFFFFFFF;
			res_patten = 0x555555;
			check_res = 0xFFFFFF;
			if (ef_write_and_compare(mode, &wdata, &rdata, data_pattern, check_data, res_patten, check_res))
			{
				return -1;
			}

			EFUSE_DBG_MSG(1, "ef_test success: mode %d\n", mode);
		}
			break;

		case 10:
		{
			check_data = data_pattern = 0x55555555;
			check_res = res_patten = 0x555555;
			if (ef_write_and_compare(mode, &wdata, &rdata, data_pattern, check_data, res_patten, check_res))
			{
				return -1;
			}

			data_pattern = 0xAAAAAAAA;
			check_data = 0xFFFFFFFF;
			res_patten = 0xAAAAAA;
			check_res = 0xFFFFFF;
			if (ef_write_and_compare(mode, &wdata, &rdata, data_pattern, check_data, res_patten, check_res))
			{
				return -1;
			}

			EFUSE_DBG_MSG(1, "ef_test success: mode %d\n", mode);
		}
			break;

		case 11:
		{
			for (bit = 0; bit < EF_TOTAL_BITS; bit++)
			{
				if (ef_get_row_col(bit, &row, &col))
					return EFUSE_STATUS_FAIL;

				memset(&wdata, 0, sizeof(EF_DATA));
				memset(&rdata, 0, sizeof(EF_DATA));
				if ((bit & 0x7F ) == 0)
					memset(tmp, 0 , sizeof(tmp));

				wdata.data[BIT_TO_WORD(row)] = (1 << (row & 0x1F));
				tmp[BIT_TO_WORD(row)] |= (1 << (row & 0x1F));

				ef_write_data(COL_TO_BYTE(col), &wdata);
				ef_read_data(COL_TO_BYTE(col), &rdata);

				for (i = 0; i < 4; i++)
				{
					if (rdata.data[i] != tmp[i])
					{
						EFUSE_DBG_MSG(1, "ef_test failed: mode %d col %x bit %d rdata.data[%x] wdata.data[%x] check_data[%x]\n",
								mode, col, bit, rdata.data[i], wdata.data[i], tmp[i]);
						return -1;
					}
				}
			}

			EFUSE_DBG_MSG(1, "ef_test success: mode %d\n", mode);
		}
			break;

	}

	return 0;
}
#endif

static int efuse_command(cmd_tbl_t *cmdtp, int flag, int argc, char * const argv[])
{
	unsigned char ldo_on =0, ldo_off = 0, sync_cnt = 0;
	unsigned short wstb_on= 0, wstb_off = 0;
	unsigned int idx = 0, data = 0;
	unsigned int mode = 0, sub = 0;

	if (argc < 2)
		return CMD_RET_USAGE;

	if (!ef_inited)
	{
		if (ef_init(0,0,0,0,0) != EFUSE_STATUS_SUCCESS)
			return EFUSE_STATUS_FAIL;

		ef_inited = 1;
	}

	if (!strncmp(argv[1], "word_read", 10))
	{
		if (argc < 3)
			return CMD_RET_USAGE;

		idx = simple_strtoul(argv[2], NULL, 16);

		printf("word_read:%d data=%x\n", ef_read_word(idx, &data), data);
	}
	else if (!strncmp(argv[1], "word_write", 11))
	{
		if (argc < 4)
			return CMD_RET_USAGE;

		idx = simple_strtoul(argv[2], NULL, 16);
		data = simple_strtoul(argv[3], NULL, 16);

		printf("word_write:%d data=%x\n", ef_write_word(idx, &data), data);
	}
#if EFUSE_TEST
	else if (!strncmp(argv[1], "test", 4))
	{
		if (argc < 4)
			return CMD_RET_USAGE;

		mode = simple_strtoul(argv[2], NULL, 10);
		sub = simple_strtoul(argv[3], NULL, 10);

		ef_test(mode, sub);
	}
#endif
	else
		return CMD_RET_USAGE;

	return 0;
}

U_BOOT_CMD(
		efuse,   7,      0,      efuse_command,
		"efuse - efuse command\n",
		"efuse usage:\n"
		"	efuse word_read [word_idx]\n"
		"	efuse word_write [word_idx] [4byte data]\n"
);

