#include <asm/types.h>
#include <asm/io.h>
#include <asm/tc3162.h>
#include "newspiflash.h"

static char module_name[] = "spiflash";

#define MIN(a,b)        ((a) < (b) ? (a) : (b))
#define FALSE 	0
#define TRUE 	1

#define ID_MASK         		0xffff

#define MANUFACTURER_ID(id) 	((id >> 16) & ID_MASK)
#define DEVICE_ID(id)       	(id & ID_MASK)

#define SIZE_64KiB      		0x10000
#define SIZE_64MiB              0x4000000
#define SIZE_32MiB              0x2000000
#define SIZE_16MiB              0x1000000
#define SIZE_8MiB       		0x800000
#define SIZE_4MiB       		0x400000
#define SIZE_2MiB       		0x200000

/* Manufacturers */
#define MANUFACTURER_ST 		0x0020
#define MANUFACTURER_WINBOND	0x00ef
#define MANUFACTURER_SST		0x00bf
#define MANUFACTURER_MXIC		0x00c2
#define MANUFACTURER_SPANSION	0x0001
#define MANUFACTURER_EON		0x001c
#define MANUFACTURER_NUMONYX	0x0020
#define MANUFACTURER_GIGADEVICE	0x00c8
#define MANUFACTURER_ESMT		0x008c

/* ST */
#define M25P16	          		0x2015
#define M25P32	          		0x2016
#define M25P64	          		0x2017

/* Winbond */
#define W25X16					0x3015
#define W25X32					0x3016
#define W25X64					0x3017
#define W25X128					0x3018
#define W25X256					0x3019

#define W25Q16					0x4015
#define W25Q32					0x4016
#define W25Q64					0x4017
#define W25Q128					0x4018
#define W25Q256					0x4019

/* SST */
#define SST25VF032B				0x254a

/* MXIC */
#define MX25L16					0x2015
#define MX25L3205D				0x2016
#define MX25L6405D				0x2017
#define MX25L12805D				0x2018
#define MX25L25635E  			0x2019

/* SPANSION */
#define S25FL016A          		0x0214
#define S25FL032A          		0x0215
#define S25FL064A          		0x0216
#define S25FL128P          		0x2018
#define S25FL256S				0x0219
#define S25FL016K          		0x4015

#define S25FL208K          		0x4014
#define S25FL116K          		0x4015
#define S25FL216K          		0x4015
#define S25FL127S         		0x2018

/* MICRON(NUMONYX) */
#define N25Q064					0xba17
#define N25Q128					0xba18
#define N25Q256					0xba19

/* EON */
#define EN25Q64          		0x3017

/* GIGA */
#define GD25Q16				0x4015
#define GD25Q32				0x4016
#define GD25Q64				0x4017
#define GD25Q128B			0x4018
#define GD25Q256C			0x4019

/* ESMT */
#define F25L16PA			0x2115
#define F25L32QA			0x4116
#define F25L64QA			0x4117


/* define SPI mode */
#define SPI_STD_RD				0x0
#define SPI_STD_FAST_RD			0x1
#define SPI_FAST_RD_DUAL_O		0x2
#define SPI_FAST_RD_DUAL_IO		0x3
#define SPI_FAST_RD_QUAD_O		0x4
#define SPI_FAST_RD_QUAD_IO		0x5
#define SPI_BURST_RD_QUAD_IO	0x6

#define SPI_REG_BASE		0x1fbc0000
#define SPI_REG_MASTER		0x1fbc0028
#define SPI_REG_MOREBUF		0x1fbc002c
#define SPI_REG_SPACE_CR	0x1fbc003c
#define SPI_FLASH_DATA2          0x0C
#define SPI_FLASH_DATA3          0x10
#define SPI_FLASH_DATA4          0x14
#define SPI_FLASH_DATA5          0x18
#define SPI_FLASH_DATA6          0x1C
#define SPI_FLASH_DATA7          0x20
#define SPI_FLASH_DATA8          0x24
struct spi_flash_info {
	const u16 mfr_id;
	const u16 dev_id;
	const u16 extra_id;
	const char *name;
	const int DeviceSize;
	const int EraseSize;
	int mode;
};

#define printf	prom_printf

struct mtd_info {
	unsigned long offset; /* Offset within the memory */
	unsigned long size;   /* Total size of the MTD */
	unsigned long erasesize;
};

int spiflash_init (unsigned long rom_base);
void spiflash_exit (void);
int flash_index;
static int spiflash_probe (void);
static int spiflash_erase_internal (struct mtd_info *mtd,unsigned long addr, unsigned long len);
static int spiflash_read_internal (struct mtd_info *mtd, unsigned long from,unsigned long len,unsigned long *retlen,unsigned char *buf);
static int spiflash_write_internal (struct mtd_info *mtd,unsigned long to,unsigned long len,unsigned long *retlen,const unsigned char *buf, int speed);

/* Mapping of generic opcodes to STM serial flash opcodes */

struct mtd_info mtd;
int program_speed = PAGE_PROGRAM;

unsigned long spidbg = 0;
static __u8 byte_program_mode = 0;
static __u8 is_W25Q256 = 0;
static __u8 is_MX25L256 = 0;
static __u8 is_S25FL256S = 0;
static __u8 is_N25Q256 = 0;

void send_single_opcode_cmd(uint8 op_code)
{
	SEND_CSL_CMD;
	SEND_OPFIFO_WRITE_CMD(OP_OUTS, OP_REPEAT_ONE);
	SEND_DFIFO_WRITE_CMD(op_code);
	SEND_CSH_CMD;
}

void sf_manual_begin(void)
{	
	/* disable READ_IDLE_EN */
	SEND_READ_IDLE_DIS_CMD;
	/* wait until auto read status is IDLE */
	while(ReadReg(SF_RDCTL_FSM)) ;
	/* auto mode -> manual mode */
	SEND_AUTO2MANUAL_CMD;
}

void sf_manual_end(void) 
{
	/* manual mode -> auto mode */
	SEND_MANUAL2AUTO_CMD;
	/* enable READ_IDLE_EN */
	SEND_READ_IDLE_EN_CMD;
}

void sf_manual_wren_begin(void)
{	
	sf_manual_begin();
	SEND_WREN_CMD;
}

void sf_manual_wren_end(void) 
{
	uint8 reg_sr1 = 0;
	/* wait until WIP bit is idle */
	SEND_WAIT_WIP_IDLE_INTERNAL_CMD(reg_sr1);
	
	sf_manual_end();
}

void enter_4Byte_mode(void)
{
	if(is_S25FL256S){
		SEND_WRITE_REGISTER_CMD(SF_OP_WR_BANKREG, SF_BANKREG_4B);
		SEND_WRDIS_CMD; // spansion bank register does not require the WREN command to precede it, which means WREN bit need to disable by manually.
	} else if(is_N25Q256){
		SEND_ENTER_4BYTES_MODE_WREN_CMD; //for N25Q256, to enter/exit 4Byte address mode, the WREN bit must be set to 1 first.
	}else{
		SEND_ENTER_4BYTES_MODE_CMD;
	}
}

static u32
spiflash_read_id (void)
{
	u32 flash_id=0;
	u8 id[3]={0};
	u8 cur_mode=0;
	
	SEND_RD_ID_CMD(id);
	
	flash_id = id[0] << 16 | id[1] << 8 | id[2];

	return flash_id;
}

static int
spiflash_erase_oneblock (struct mtd_info *mtd, unsigned long addr, unsigned long len)
{
	u8 cur_mode=0;
	u8 mode=0;

   	/* sanity checks */
   	if (addr + len > mtd->size) return (-1);

	if(mtd->erasesize == SIZE_64KiB) {
		if(mtd->size >= SIZE_32MiB)
		{
			mode = 1;
		} else {
			mode = 0;
		}
		SEND_SECTOR_ERASE_CMD(addr, mode);
		
		printf("\nerase addr=0x%08x", addr);
	} else {
		return (-1);
	}
	
   	return (0);
}

static int
spiflash_erase_internal (struct mtd_info *mtd, unsigned long addr, unsigned long len)
{
	int ret=0;

	if (addr & (mtd->erasesize - 1))
	  	addr = addr - (addr & (mtd->erasesize - 1));
	if (len & (mtd->erasesize - 1))
		len = len - (len & (mtd->erasesize - 1)) + mtd->erasesize;

	if(len + addr > mtd->size)
		return -1;

	while (len) {
		ret = spiflash_erase_oneblock(mtd, addr, len);
		if (ret < 0) return ret;

		addr += mtd->erasesize;
		len -= mtd->erasesize;
	}

	return 0;
}

static int
spiflash_manual_read_internal (struct mtd_info *mtd, unsigned long from,unsigned long len,unsigned long *retlen,unsigned char *buf, unsigned char read_type)
{
	int mode=0;
	u8 cur_mode=0;
	unsigned long dest_addr = from;

   	/* sanity checks */
   	if (!len) return (0);
   	if (from + len > mtd->size) return (-1);

	if(mtd->size >= SIZE_32MiB) {
		mode = 1;
	} else {
		mode = 0;
	}
		switch(read_type){
		case READ_DATA:
			SEND_READ_CMD(from, buf, len, mode);
			break;
			
		case FAST_READ:
			SEND_FASTREAD_CMD(from, buf, len, mode);
			break;
			
		case FAST_READ_DUALOUT:
			SEND_FASTREAD_DUALOUT_CMD(from, buf, len, mode);
			break;
			
		case FAST_READ_DUALIO:
			SEND_FASTREAD_DUALIO_CMD(from, buf, len, mode);
			break;

		case FAST_READ_QUADOUT:
			SEND_FASTREAD_QUADOUT_CMD(from, buf, len, mode);
			break;

		case FAST_READ_QUADIO:
			SEND_FASTREAD_QUADIO_CMD(from, buf, len, mode);
			break;

		default:
			printf("spiflash_read: not support this mode\n");
			break;		
		}
   	*retlen = len;

   	return (0);
}

static int
spiflash_manual_read_data(struct mtd_info *mtd, unsigned long from,unsigned long len,unsigned long *retlen,unsigned char *buf)
{
	spiflash_manual_read_internal(mtd, from, len, retlen, buf, READ_DATA);
   	return (0);
}

static int
spiflash_manual_fast_read(struct mtd_info *mtd, unsigned long from,unsigned long len,unsigned long *retlen,unsigned char *buf)
{
	spiflash_manual_read_internal(mtd, from, len, retlen, buf, FAST_READ);
   	return (0);
}

static int
spiflash_manual_fast_read_dualOut(struct mtd_info *mtd, unsigned long from,unsigned long len,unsigned long *retlen,unsigned char *buf)
{
	spiflash_manual_read_internal(mtd, from, len, retlen, buf, FAST_READ_DUALOUT);
   	return (0);
}

static int
spiflash_manual_fast_read_dualIO(struct mtd_info *mtd, unsigned long from,unsigned long len,unsigned long *retlen,unsigned char *buf)
{
	spiflash_manual_read_internal(mtd, from, len, retlen, buf, FAST_READ_DUALIO);
   	return (0);
}

static int
spiflash_manual_fast_read_quadOut(struct mtd_info *mtd, unsigned long from,unsigned long len,unsigned long *retlen,unsigned char *buf)
{
	spiflash_manual_read_internal(mtd, from, len, retlen, buf, FAST_READ_QUADOUT);
   	return (0);
}

static int
spiflash_manual_fast_read_quadIO(struct mtd_info *mtd, unsigned long from,unsigned long len,unsigned long *retlen,unsigned char *buf)
{
	spiflash_manual_read_internal(mtd, from, len, retlen, buf, FAST_READ_QUADIO);
   	return (0);
}

static int
spiflash_read_internal (struct mtd_info *mtd, unsigned long from,unsigned long len,unsigned long *retlen,unsigned char *buf)
{
	u8 cur_mode=0;
	
   	/* sanity checks */
   	if (!len) return (0);
   	if (from + len > mtd->size)	return (-1);
	
   	*retlen = len;
	SEND_READ_IDLE_EN_CMD;

	memcpy4(buf, (unsigned char*)((mtd->offset + from)), len);

   	return (0);
}

static int
spiflash_write_internal (struct mtd_info *mtd,unsigned long to,unsigned long len,unsigned long *retlen,const unsigned char *buf, int program_speed)
{
	u8 cur_mode=0;
	u8 mode=0;
	static int cnt=0;
	u8 reg_sr3 = 0;
	unsigned long pp_len = 0;
	unsigned long pp_cur_len = 0;
	unsigned long tmp_len = 0;
	u32 flags;

   	*retlen = 0;

   	/* sanity checks */
   	if (!len) return (0);
   	if (to + len > mtd->size) return (-1);

	//printf("program from %x to %x\n", to, to+len);

	pp_len = len;

	if(mtd->size >= SIZE_32MiB) {
		mode = 1;
	}
	else {
		mode = 0;
	}

	while(pp_len > 0) {
		cnt++;
		if ((cnt & 0x3f) == 0)
			printf(".");
		
		if((to & (SF_PP_MAX-1)) != 0) {
			pp_cur_len = MIN((SF_PP_MAX - (to % SF_PP_MAX)), pp_len);
		}
		else {
			pp_cur_len = MIN(SF_PP_MAX, pp_len);
		}
		pp_len -= pp_cur_len;
		
		tmp_len = pp_cur_len;

		if(program_speed == PAGE_PROGRAM) {
			SEND_PAGE_PROGRAM_CMD(to, buf, pp_cur_len, mode);
		} else {
			SEND_PAGE_QUAD_PROGRAM_CMD(to, buf, pp_cur_len, mode);
		}
		
		to += tmp_len;
	}

   	return (0);
}

static struct spi_flash_info flash_tables[] = {
	{
		mfr_id: MANUFACTURER_ST,
		dev_id: M25P64,
		name: "ST M25P64",
		DeviceSize: SIZE_8MiB,
		EraseSize: SIZE_64KiB,
		mode: SPI_STD_FAST_RD,
	},
	{
		mfr_id: MANUFACTURER_ST,
		dev_id: M25P32,
		name: "ST M25P32",
		DeviceSize: SIZE_4MiB,
		EraseSize: SIZE_64KiB,
		mode: SPI_STD_FAST_RD,
	},
	{
		mfr_id: MANUFACTURER_ST,
		dev_id: M25P16,
		name: "ST M25P16",
		DeviceSize: SIZE_2MiB,
		EraseSize: SIZE_64KiB,
		mode: SPI_STD_FAST_RD,
	},
	{
		mfr_id: MANUFACTURER_WINBOND,
		dev_id: W25X256,
		name: "Winbond W25X256",
		DeviceSize: SIZE_32MiB,
		EraseSize: SIZE_64KiB,
		mode: SPI_FAST_RD_DUAL_O,
	},
	{
		mfr_id: MANUFACTURER_WINBOND,
		dev_id: W25X128,
		name: "Winbond W25X128",
		DeviceSize: SIZE_16MiB,
		EraseSize: SIZE_64KiB,
		mode: SPI_FAST_RD_DUAL_O,
	},
	{
		mfr_id: MANUFACTURER_WINBOND,
		dev_id: W25X64,
		name: "Winbond W25X64",
		DeviceSize: SIZE_8MiB,
		EraseSize: SIZE_64KiB,
		mode: SPI_FAST_RD_DUAL_O,
	},
	{
		mfr_id: MANUFACTURER_WINBOND,
		dev_id: W25X32,
		name: "Winbond W25X32",
		DeviceSize: SIZE_4MiB,
		EraseSize: SIZE_64KiB,
		mode: SPI_FAST_RD_DUAL_O,
	},
	{
		mfr_id: MANUFACTURER_WINBOND,
		dev_id: W25X16,
		name: "Winbond W25X16",
		DeviceSize: SIZE_2MiB,
		EraseSize: SIZE_64KiB,
		mode: SPI_FAST_RD_DUAL_O,
	},
	{
		mfr_id: MANUFACTURER_WINBOND,
		dev_id: W25Q256,
		name: "Winbond W25Q256",
		DeviceSize: SIZE_32MiB,
		EraseSize: SIZE_64KiB,
		mode: SPI_FAST_RD_QUAD_IO,
	},
	{
		mfr_id: MANUFACTURER_WINBOND,
		dev_id: W25Q128,
		name: "Winbond W25Q128",
		DeviceSize: SIZE_16MiB,
		EraseSize: SIZE_64KiB,
		mode: SPI_FAST_RD_QUAD_IO,
	},
	{
		mfr_id: MANUFACTURER_WINBOND,
		dev_id: W25Q64,
		name: "Winbond W25Q64",
		DeviceSize: SIZE_8MiB,
		EraseSize: SIZE_64KiB,
		mode: SPI_FAST_RD_QUAD_IO,
	},
	{
		mfr_id: MANUFACTURER_WINBOND,
		dev_id: W25Q32,
		name: "Winbond W25Q32",
		DeviceSize: SIZE_4MiB,
		EraseSize: SIZE_64KiB,
		mode: SPI_FAST_RD_QUAD_IO,
	},
	{
		mfr_id: MANUFACTURER_WINBOND,
		dev_id: W25Q16,
		name: "Winbond W25Q16",
		DeviceSize: SIZE_2MiB,
		EraseSize: SIZE_64KiB,
		mode: SPI_FAST_RD_QUAD_IO,
	},
	{
		mfr_id: MANUFACTURER_SST,
		dev_id: SST25VF032B,
		name: "SST 25VF032B",
		DeviceSize: SIZE_4MiB,
		EraseSize: SIZE_64KiB,
		mode: SPI_STD_FAST_RD,
	},
	{
		mfr_id: MANUFACTURER_MXIC,
		dev_id: MX25L16,
		name: "MX25L16",
		DeviceSize: SIZE_2MiB,
		EraseSize: SIZE_64KiB,
		mode: SPI_STD_FAST_RD,
	},
	{
		mfr_id: MANUFACTURER_MXIC,
		dev_id: MX25L3205D,
		name: "MX25L3205D",
		DeviceSize: SIZE_4MiB,
		EraseSize: SIZE_64KiB,
		mode: SPI_STD_FAST_RD,
	},
	{
		mfr_id: MANUFACTURER_MXIC,
		dev_id: MX25L6405D,
		name: "MX25L6405D",
		DeviceSize: SIZE_8MiB,
		EraseSize: SIZE_64KiB,
		mode: SPI_STD_FAST_RD,
	},
	{
		mfr_id: MANUFACTURER_MXIC,
		dev_id: MX25L12805D,
		name: "MX25L12805D",
		DeviceSize: SIZE_16MiB,
		EraseSize: SIZE_64KiB,
		mode: SPI_STD_FAST_RD,
	},
	{
		mfr_id: MANUFACTURER_MXIC,
		dev_id: MX25L25635E,
		name: "MX25L25635E",
		DeviceSize: SIZE_32MiB,
		EraseSize: SIZE_64KiB,
		mode: SPI_FAST_RD_QUAD_IO,
	},
	{
		mfr_id: MANUFACTURER_SPANSION,
		dev_id: S25FL064A,
		name: "S25FL064A",
		DeviceSize: SIZE_8MiB,
		EraseSize: SIZE_64KiB,
		mode: SPI_STD_FAST_RD,
	},
	{
		mfr_id: MANUFACTURER_SPANSION,
		dev_id: S25FL032A,
		name: "S25FL032A",
		DeviceSize: SIZE_4MiB,
		EraseSize: SIZE_64KiB,
		mode: SPI_STD_FAST_RD,
	},
	{
		mfr_id: MANUFACTURER_SPANSION,
		dev_id: S25FL016A,
		name: "S25FL016A",
		DeviceSize: SIZE_2MiB,
		EraseSize: SIZE_64KiB,
		mode: SPI_STD_FAST_RD,
	},
	{
		mfr_id: MANUFACTURER_SPANSION,
		dev_id: S25FL128P,
		name: "S25FL128P",
		DeviceSize: SIZE_16MiB,
		EraseSize: SIZE_64KiB,
		mode: SPI_STD_FAST_RD,
	},
	{
		mfr_id: MANUFACTURER_SPANSION,
		dev_id: S25FL256S,
		name: "Spasion S25FL256S",
		DeviceSize: SIZE_32MiB,
		EraseSize: SIZE_64KiB,
		mode: SPI_STD_FAST_RD,
	},
	{
		mfr_id: MANUFACTURER_EON,
		dev_id: EN25Q64,
		name: "EN25Q64",
		DeviceSize: SIZE_8MiB,
		EraseSize: SIZE_64KiB,
		mode: SPI_FAST_RD_QUAD_IO,
	},
	{
		mfr_id: MANUFACTURER_NUMONYX,
		dev_id: N25Q064,
		name: "N25Q064",
		DeviceSize: SIZE_8MiB,
		EraseSize: SIZE_64KiB,
		mode: SPI_STD_RD,
	},
	{
		mfr_id: MANUFACTURER_NUMONYX,
		dev_id: N25Q128,
		name: "N25Q128",
		DeviceSize: SIZE_16MiB,
		EraseSize: SIZE_64KiB,
		mode: SPI_STD_RD,
	},
	{
		mfr_id: MANUFACTURER_NUMONYX,
		dev_id: N25Q256,
		name: "N25Q256",
		DeviceSize: SIZE_32MiB,
		EraseSize: SIZE_64KiB,
		mode: SPI_STD_RD,
	},
	{
		mfr_id: MANUFACTURER_SPANSION,
		dev_id: S25FL016K,
		name: "S25FL016K",
		DeviceSize: SIZE_2MiB,
		EraseSize: SIZE_64KiB,
		mode: SPI_FAST_RD_DUAL_IO,
	},
	{
		mfr_id: MANUFACTURER_GIGADEVICE,
		dev_id: GD25Q16,
		name: "GD25Q16",
		DeviceSize: SIZE_2MiB,
		EraseSize: SIZE_64KiB,
		mode: SPI_FAST_RD_DUAL_IO,
	},
	{
		mfr_id: MANUFACTURER_GIGADEVICE,
		dev_id: GD25Q32,
		name: "GD25Q32",
		DeviceSize: SIZE_4MiB,
		EraseSize: SIZE_64KiB,
		mode: SPI_FAST_RD_DUAL_IO,
	},
	{
		mfr_id: MANUFACTURER_GIGADEVICE,
		dev_id: GD25Q64,
		name: "GD25Q64",
		DeviceSize: SIZE_8MiB,
		EraseSize: SIZE_64KiB,
		mode: SPI_FAST_RD_DUAL_IO,
	},
	{
		mfr_id: MANUFACTURER_GIGADEVICE,
		dev_id: GD25Q128B,
		name: "GD25Q128B",
		DeviceSize: SIZE_16MiB,
		EraseSize: SIZE_64KiB,
		mode: SPI_FAST_RD_DUAL_IO,
	},
	{
		mfr_id: MANUFACTURER_GIGADEVICE,
		dev_id: GD25Q256C,
		name: "GD25Q256C",
		DeviceSize: SIZE_32MiB,
		EraseSize: SIZE_64KiB,
		mode: SPI_FAST_RD_DUAL_IO,
	},
	{
		mfr_id: MANUFACTURER_ESMT,
		dev_id: F25L16PA,
		name: "F25L16PA",
		DeviceSize: SIZE_2MiB,
		EraseSize: SIZE_64KiB,
		mode: SPI_FAST_RD_DUAL_IO,
	},
	{
		mfr_id: MANUFACTURER_ESMT,
		dev_id: F25L32QA,
		name: "F25L32QA",
		DeviceSize: SIZE_4MiB,
		EraseSize: SIZE_64KiB,
		mode: SPI_FAST_RD_DUAL_IO,
	},
	{
		mfr_id: MANUFACTURER_ESMT,
		dev_id: F25L64QA,
		name: "F25L64QA",
		DeviceSize: SIZE_8MiB,
		EraseSize: SIZE_64KiB,
		mode: SPI_FAST_RD_DUAL_IO,
	},
	{
		mfr_id: MANUFACTURER_SPANSION,
		dev_id: S25FL208K,
		name: "S25FL208K",
		DeviceSize: SIZE_2MiB,
		EraseSize: SIZE_64KiB,
		mode: SPI_STD_FAST_RD,
	},
	{
		mfr_id: MANUFACTURER_SPANSION,
		dev_id: S25FL116K,
		name: "S25FL116K",
		DeviceSize: SIZE_2MiB,
		EraseSize: SIZE_64KiB,
		mode: SPI_STD_FAST_RD,
	},
	{
		mfr_id: MANUFACTURER_SPANSION,
		dev_id: S25FL216K,
		name: "S25FL216K",
		DeviceSize: SIZE_2MiB,
		EraseSize: SIZE_64KiB,
		mode: SPI_STD_FAST_RD,
	},
	{
		mfr_id: MANUFACTURER_SPANSION,
		dev_id: S25FL127S,
		name: "S25FL127S",
		DeviceSize: SIZE_16MiB,
		EraseSize: SIZE_64KiB,
		mode: SPI_STD_FAST_RD,
	},
};


static int spiflash_probe(void)
{
	int i=0;
	unsigned long flash_id=0;

	flash_id = spiflash_read_id();

	for (i=0; i < (sizeof(flash_tables)/sizeof(struct spi_flash_info)); i++) {
		if ((MANUFACTURER_ID(flash_id) == flash_tables[i].mfr_id) &&
		    (DEVICE_ID(flash_id) == flash_tables[i].dev_id)) {
			if ( (MANUFACTURER_ID(flash_id) == MANUFACTURER_WINBOND) && (DEVICE_ID(flash_id) == W25Q256) ) {
				is_W25Q256 = 1;
			}
			if ( (MANUFACTURER_ID(flash_id) == MANUFACTURER_MXIC) && (DEVICE_ID(flash_id) == MX25L25635E) ) {
				is_MX25L256 = 1;
			}
			if ( (MANUFACTURER_ID(flash_id) == MANUFACTURER_SPANSION) && (DEVICE_ID(flash_id) == S25FL256S) ) {
				is_S25FL256S = 1;
			}
			if ( (MANUFACTURER_ID(flash_id) == MANUFACTURER_NUMONYX) && (DEVICE_ID(flash_id) == N25Q256) ) {
				is_N25Q256 = 1;
			}
			return i;
		}
	}
	return -1;
}

int spiflash_erase(unsigned long addr, unsigned long size)
{
	return spiflash_erase_internal(&mtd, addr, size);
}

int spiflash_read(unsigned long from,
	unsigned long len, unsigned long *retlen, unsigned char *buf)
{
	return spiflash_manual_fast_read(&mtd, from, len, retlen, buf);
}

int spiflash_write(unsigned long to,
	unsigned long len, unsigned long *retlen, const unsigned char *buf)
{
	return spiflash_write_internal(&mtd, to, len, retlen, buf, program_speed);
}

unsigned char ReadSPICache(unsigned long index)
{
	unsigned long retlen = 0;
	unsigned long len = 1;
	unsigned char buf[2]={0};
	index &= 0x03FFFFFF;
	spiflash_read_internal(&mtd, index, len, &retlen, buf);
	return buf[0];
}

unsigned char ReadSPIByte(unsigned long index)
{
	unsigned long retlen = 0;
	unsigned long len = 1;
	unsigned char buf[2]={0};
	index &= 0x03FFFFFF;
	spiflash_read_internal(&mtd, index, len, &retlen, buf);
	return buf[0];
}

unsigned long ReadSPIDWord(unsigned long index)
{
	unsigned long retlen = 0;
	unsigned long dword=0;
	
	index &= 0x03FFFFFF;

	spiflash_read_internal(&mtd, index, 4, &retlen, &dword);
	
	return dword;
}

/* only use in boot.img */
void Quad_Mode_Enable(int index)
{
	u8 val=0;
	u8 check=0;

	if( (flash_tables[index].mfr_id == MANUFACTURER_MXIC) )
	{
		printf("\nMXIC Quad Mode Enable\n");
		SEND_READ_REGISTER_CMD(SF_OP_RD_STATUS1, val );
		val |= 0x40;
		SEND_WRITE_REGISTER_CMD(SF_OP_WR_STATUS1, val);

		SEND_READ_REGISTER_CMD(SF_OP_RD_STATUS1, check );
		printf("Status Register 1 = 0x%x\n", check);

		
	}

	if( (flash_tables[index].mfr_id == MANUFACTURER_WINBOND) )
	{

		if( (flash_tables[index].dev_id == W25Q64) )
		{
			printf("\nWinbond Quad Mode Enable\n");
			SEND_READ_REGISTER_CMD(SF_OP_RD_STATUS2, val );
			val |= 0x2;
			
			SEND_WRITE_REGISTER_TWO_BYTE_CMD(SF_OP_WR_STATUS1, 0x0, val);

			SEND_READ_REGISTER_CMD(SF_OP_RD_STATUS2, check );
			printf("Status Register 2 = 0x%x\n", check);
		}
		else
		{
			printf("\nWinbond Quad Mode Enable\n");
			SEND_READ_REGISTER_CMD(SF_OP_RD_STATUS2, val );
			val |= 0x2;
			SEND_WRITE_REGISTER_CMD(SF_OP_WR_STATUS2, val);

			SEND_READ_REGISTER_CMD(SF_OP_RD_STATUS2, check );
			printf("Status Register 2 = 0x%x\n", check);		
		}
	}


	if( (flash_tables[index].mfr_id == MANUFACTURER_GIGADEVICE) )
	{
		printf("\nGigaDevice Quad Mode Enable\n");
		if( (flash_tables[index].dev_id == GD25Q256C) ) {
			SEND_READ_REGISTER_CMD(SF_OP_RD_STATUS1, val );
			val |= 0x40;
			SEND_WRITE_REGISTER_CMD(SF_OP_WR_STATUS1, val);

			SEND_READ_REGISTER_CMD(SF_OP_RD_STATUS1, check );
			printf("Status Register 1 = 0x%x\n", check);
		}
		else {
			SEND_READ_REGISTER_CMD(SF_OP_RD_STATUS2, val );
			val |= 0x2;
			SEND_WRITE_REGISTER_CMD(SF_OP_WR_STATUS2, val);

			SEND_READ_REGISTER_CMD(SF_OP_RD_STATUS2, check );
			printf("Status Register 2 = 0x%x\n", check);
		}
	}	
}

/*------------------------------------------------------------------------------------
 * FUNCTION: spi_nor_set_clock_speed( u32 clock_factor)
 * PURPOSE : To set SPI clock. 
 *                 clock_factor = 0
 *                   EN7512: SPI clock = 500MHz / 40 = 12.5MHz
 *                   EN7522: SPI clock = 400MHz / 40 = 10MHz
 *                 clock_factor > 0
 *                   EN7512: SPI clock = 500MHz / (clock_factor * 2)
 *                   EN7522: SPI clock = 400MHz / (clock_factor * 2)
 * CALLED BY
 *   -
 * CALLS
 *   -
 * PARAMs  :
 *   INPUT : clock_factor - The SPI clock divider.
 * RETURN  : NONE.
 * NOTES   :
 * MODIFICTION HISTORY:
 *
 *------------------------------------------------------------------------------------
 */
void spi_nor_set_clock_speed( u32 clk)
{
	u32 val;
	u32 dividend;
	u32 clock_factor;

	if(!isFPGA) {
		dividend = 500;

		clock_factor = (dividend / (clk * 2));
		
		val  = VPint(SF_CLK_CHANGE);
		val &= 0xffff0000;
		VPint(SF_CLK_CHANGE) = val;

		val |= (((clock_factor) << 8)|1);
		VPint(SF_CLK_CHANGE) = val;

		printf("Set SPI Clock to %u Mhz\n", (dividend / (clock_factor * 2)));
	} else {
		printf("FPGA can not Set SPI Clock, FPGA SPI Clock is fixed to 25 Mhz\n");
	}
}
int
spiflash_init (unsigned long rom_base)
{
   	int index=0;
	int spi_read_bytes=0;
	unsigned int val;
	int	spi_controller_clk;

	/* turn off flash write protect */
	//tc_outl(CR_GPIO_DATA, tc_inl(CR_GPIO_DATA) | (1 << 5));

	mtd.offset = rom_base;
	mtd.size = 0;

	/* enable SF_MAXMUX_SEL */
	WriteReg(SF_MACMUX_SEL, MANUAL_MACMUXSEL);
	
   	index = spiflash_probe();
	if (index < 0) {
    	printf ("%s: Found no SPI flash device\n", module_name);
    	return -1;
   	}
	
	if(flash_tables[index].DeviceSize >= SIZE_32MiB){
		//enter_dualIO_enable_mode();
		enter_4Byte_mode();
		SEND_AUTO_4B_CMD;
	} else {
		SEND_AUTO_3B_CMD;
	}
	SEND_AUTO_READ_CMD;
	
	SEND_WRITE_REGISTER_CMD(SF_OP_WR_STATUS1, SF_SR1_DEFAULT);

	/* Quad_Mode_Enable(index); */

	// set SFC Clock to 25MHZ
	spi_controller_clk = 25;
	spi_nor_set_clock_speed(spi_controller_clk);

	// set Tn to 1
	WriteReg(SF_SI_CK_SEL, 9);
	
		switch (flash_tables[index].mode) {
			case SPI_FAST_RD_QUAD_O:
				flash_tables[index].mode = SPI_FAST_RD_DUAL_O;
				break;
			case SPI_FAST_RD_QUAD_IO:
				flash_tables[index].mode = SPI_FAST_RD_DUAL_IO;
				break;
			case SPI_BURST_RD_QUAD_IO:
				flash_tables[index].mode = SPI_FAST_RD_DUAL_IO;
				break;
		}

	printf("Found SPI Flash %dMiB %s at 0x%x\r\n", 
		(flash_tables[index].DeviceSize)/(1024*1024), flash_tables[index].name, rom_base);

   	mtd.size = flash_tables[index].DeviceSize;
   	mtd.erasesize = flash_tables[index].EraseSize;
   	return 0;
}

